from __future__ import annotations

import shlex
from abc import ABC, abstractmethod
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import InitVar, dataclass
from enum import Enum
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import ClassVar, Optional, Type

from dcqc.file import File
from dcqc.mixins import SerializableMixin, SerializedObject
from dcqc.target import Target


class TestStatus(Enum):
    NONE = "pending"
    FAIL = "failed"
    PASS = "passed"
    SKIP = "skipped"


# TODO: Look into the @typing.final decorator
class TestABC(SerializableMixin, ABC):
    """Abstract base class for QC tests.

    Args:
        target (Target): Single- or multi-file target.
        skip (bool, optional): Whether to skip this test,
            resulting in ``TestStatus.SKIP`` as status.
            Defaults to False.

    Raises:
        ValueError: If the test expects a single file and
            the given target features multiple files.
    """

    # Class attributes
    tier: ClassVar[int]
    is_external_test: ClassVar[bool]
    is_external_test = False
    only_one_file_targets: ClassVar[bool]
    only_one_file_targets = True

    # Instance attributes
    type: str
    target: Target

    def __init__(self, target: Target, skip: bool = False):
        self.type = self.__class__.__name__
        self.target = target
        self._status = TestStatus.SKIP if skip else TestStatus.NONE

        files = self.target.files
        if self.only_one_file_targets and len(files) > 1:
            message = f"Test ({self.type}) expected one file, not multiple ({files})."
            raise ValueError(message)

    def skip(self):
        """Force the test to be skipped."""
        self._status = TestStatus.SKIP

    def get_status(self, compute_ok: bool = True) -> TestStatus:
        """Compute (if applicable) and return the test status."""
        if self._status == TestStatus.NONE and compute_ok:
            self._status = self.compute_status()
        return self._status

    def get_files(self, staged: bool = True) -> list[File]:
        """Get and stage files for target.

        Args:
            staged: Whether to make sure that the files are staged.
            Defaults to True.

        Returns:
            Staged target files.
        """
        files = []
        for file in self.target.files:
            if staged:
                file.stage()
            files.append(file)
        return files

    def get_file(self, staged: bool = True) -> File:
        """Get and stage file for single-file target.

        Args:
            staged: Whether to make sure that the files are staged.
            Defaults to True.

        Raises:
            ValueError: If the target has multiple files.

        Returns:
            Staged target file.
        """
        files = self.get_files(staged)
        if len(files) != 1:
            message = "This method only supports single-file targets."
            raise ValueError(message)
        return files[0]

    @classmethod
    def get_subclass_by_name(cls, test: str) -> Type[TestABC]:
        """Retrieve subclass by name."""
        test_classes = TestABC.__subclasses__()
        registry = {test_class.__name__: test_class for test_class in test_classes}
        if test not in registry:
            test_names = list(registry)
            message = f"Test ({test}) not among available options ({test_names})."
            raise ValueError(message)
        return registry[test]

    @classmethod
    def list_subclasses(cls) -> list[Type[TestABC]]:
        """List all subclasses."""
        test_classes = TestABC.__subclasses__()
        return test_classes

    @abstractmethod
    def compute_status(self) -> TestStatus:
        """Compute the status of the test."""

    def to_dict(self) -> SerializedObject:
        test_dict = {
            "type": self.type,
            "tier": self.tier,
            "is_external_test": self.is_external_test,
            "status": self._status.value,
            "target": self.target.to_dict(),
        }
        return test_dict

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> TestABC:
        """Deserialize a dictionary into a test.

        Args:
            dictionary: A serialized test object.

        Returns:
            The reconstructed test object.
        """
        test_cls_name = dictionary.pop("type")
        test_cls = cls.get_subclass_by_name(test_cls_name)

        target_dict = dictionary["target"]
        target = Target.from_dict(target_dict)

        test = test_cls(target)

        status = TestStatus(dictionary["status"])
        test._status = status

        return test

    def import_module(self, name: str) -> ModuleType:
        try:
            module = import_module(name)
        except ModuleNotFoundError:
            message = (
                f"{self.type} cannot be computed without the '{name}' package. ",
                "Re-install `dcqc` with the `all` extra: pip install dcqc[all].",
            )
            raise ModuleNotFoundError(message)
        return module


@dataclass
class Process(SerializableMixin):
    container: str
    command_args: InitVar[Sequence[str]]
    cpus: int = 1
    memory: int = 2  # In GB

    _serialized_properties = ["command"]

    def __post_init__(self, command_args: Sequence[str]):
        self._command_args = command_args

    @property
    def command(self) -> str:
        return shlex.join(self._command_args)

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> Process:
        """Deserialize a dictionary into a process.

        Args:
            dictionary: A serialized proces object.

        Returns:
            The reconstructed process object.
        """
        dictionary = deepcopy(dictionary)
        command = dictionary.pop("command")
        command_args = shlex.split(command)
        dictionary["command_args"] = command_args
        process = cls(**dictionary)
        return process


class ExternalTestMixin(TestABC):
    # Class attributes
    is_external_test = True

    # Class constants
    STDOUT_PATH: ClassVar[Path]
    STDOUT_PATH = Path("std_out.txt")
    STDERR_PATH: ClassVar[Path]
    STDERR_PATH = Path("std_err.txt")
    EXITCODE_PATH: ClassVar[Path]
    EXITCODE_PATH = Path("exit_code.txt")

    def compute_status(self) -> TestStatus:
        """Compute the status of the test."""
        outputs = self._find_process_outputs()
        return self._interpret_process_outputs(outputs)

    @abstractmethod
    def generate_process(self) -> Process:
        """Generate the process that needs to be run."""

    @classmethod
    def _find_process_outputs(
        cls, search_dir: Optional[Path] = None
    ) -> dict[str, Path]:
        """Locate the output files from the executed process."""
        search_dir = search_dir or Path(".")
        outputs = {
            "std_out": search_dir / cls.STDOUT_PATH,
            "std_err": search_dir / cls.STDERR_PATH,
            "exit_code": search_dir / cls.EXITCODE_PATH,
        }

        for path in outputs.values():
            if not path.exists():
                message = f"Expected process output ({path}) does not exist."
                raise FileNotFoundError(message)
        return outputs

    def _interpret_process_outputs(self, outputs: dict[str, Path]) -> TestStatus:
        """Interpret the process output files to yield a test status."""
        exit_code = outputs["exit_code"].read_text()
        exit_code = exit_code.strip()
        if exit_code == "0":
            status = TestStatus.PASS
        else:
            status = TestStatus.FAIL
        return status

    # TODO: Include process in serialized test dictionary
    # def to_dict(self):
    #     dictionary = super(ExternalTestMixin, self).to_dict()
    #     process = self.generate_process()
    #     dictionary["process"] = process.to_dict()
    #     return dictionary



import pytest

from dcqc.target import Target


def test_that_a_target_can_be_saved_and_restored_without_changing(test_files):
    test_file = test_files["good"]
    target_1 = Target(test_file)
    target_1_dict = target_1.to_dict()
    target_2 = Target.from_dict(target_1_dict)
    target_2_dict = target_2.to_dict()
    assert target_1 == target_2
    assert target_1_dict == target_2_dict


def test_for_an_error_when_restoring_a_target_with_a_discordant_type(test_files):
    test_file = test_files["good"]
    target_1 = Target(test_file)
    target_1_dict = target_1.to_dict()
    target_1_dict["type"] = "UnexpectedQcTarget"
    with pytest.raises(ValueError):
        Target.from_dict(target_1_dict)


def test_for_an_error_when_accessing_the_file_type_for_a_multifile_target(test_files):
    file = test_files["good"]
    target = Target(file, file)
    with pytest.raises(NotImplementedError):
        target.get_file_type()



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from dcqc.file import File, FileType
from dcqc.mixins import SerializableMixin, SerializedObject


# TODO: Eventually, there might be target-specific metadata
# TODO: Now that Target is much simpler, it might make sense
#       to rename the class to FileSet since it currently
#       really is just a wrapper for a group of files
# TODO: Maybe the Composite pattern would work here?
@dataclass
class Target(SerializableMixin):
    """Construct a multi-file Target.

    Targets ensure support for both single-file
    and multi-file tests.

    Args:
        *files: Sequence of files objects.
        id: A unique identifier for the target.
            Defaults to None.
    """

    type: str
    id: Optional[str]
    files: list[File]

    def __init__(self, *files: File, id: Optional[str] = None):
        self.type = self.__class__.__name__
        self.files = list(files)
        self.id = id

    def __hash__(self):
        return hash(tuple(self.files))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def get_file_type(self) -> FileType:
        """Retrieve the file type for the target.

        This function currently only supports targets
        composed of a single file.

        Raises:
            NotImplementedError: If the target has
                more or less than one file.

        Returns:
            The file type object.
        """
        num_files = len(self.files)
        if num_files == 1:
            file = self.files[0]
            file_type = file.get_file_type()
        else:
            message = f"Target has {num_files} files, which isn't supported yet."
            raise NotImplementedError(message)
        return file_type

    def stage(
        self, destination: Optional[Path] = None, overwrite: bool = False
    ) -> list[Path]:
        """Create local copy of local or remote file.

        A destination is not required for remote files; it
        defaults to a temporary directory.
        Local files aren't moved if a destination is omitted.

        Args:
            destination: File or folder where to store the file.
                Defaults to None.
            overwrite: Whether to ignore existing file at the
                target destination. Defaults to False.

        Raises:
            ValueError: If the parent directory of the
                destination does not exist.
            FileExistsError: If the destination file already
                exists and ``overwrite`` was not enabled.

        Returns:
            The path of the local copy.
        """
        paths = list()
        for file in self.files:
            path = file.stage(destination, overwrite)
            paths.append(path)
        return paths

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> Target:
        """Deserialize a dictionary into a target.

        Args:
            dictionary: A serialized target object.

        Returns:
            The reconstructed target object.
        """
        dictionary = deepcopy(dictionary)
        dictionary = cls.from_dict_prepare(dictionary)
        files = [File.from_dict(d) for d in dictionary["files"]]
        id = dictionary["id"]
        target = cls(*files, id=id)
        return target



import hashlib
import json
from pathlib import Path

from dcqc.tests.test_abc import ExternalTestMixin, Process, TestABC, TestStatus


class FileExtensionTest(TestABC):
    tier = 1
    only_one_file_targets = False

    def compute_status(self) -> TestStatus:
        status = TestStatus.PASS
        for file in self.get_files(staged=False):
            file_type = file.get_file_type()
            file_extensions = file_type.file_extensions
            if not file.name.endswith(file_extensions):
                status = TestStatus.FAIL
                break
        return status


class Md5ChecksumTest(TestABC):
    tier = 1
    only_one_file_targets = False

    def compute_status(self) -> TestStatus:
        status = TestStatus.PASS
        for file in self.get_files():
            expected_md5 = file.get_metadata("md5_checksum")
            actual_md5 = self._compute_md5_checksum(file.local_path)
            if expected_md5 != actual_md5:
                status = TestStatus.FAIL
                break
        return status

    def _compute_md5_checksum(self, path: Path) -> str:
        hash_md5 = hashlib.md5()
        with path.open("rb") as infile:
            for chunk in iter(lambda: infile.read(4096), b""):
                hash_md5.update(chunk)
        actual_md5 = hash_md5.hexdigest()
        return actual_md5


class JsonLoadTest(TestABC):
    tier = 2
    only_one_file_targets = False

    def compute_status(self) -> TestStatus:
        status = TestStatus.PASS
        for file in self.get_files():
            if not self._can_be_loaded(file.local_path):
                status = TestStatus.FAIL
                break
        return status

    def _can_be_loaded(self, path: Path) -> bool:
        success = True
        with path.open("r") as infile:
            try:
                json.load(infile)
            except Exception:
                success = False
        return success


class JsonLdLoadTest(TestABC):
    tier = 2
    only_one_file_targets = False

    def compute_status(self) -> TestStatus:
        status = TestStatus.PASS
        for file in self.get_files():
            if not self._can_be_loaded(file.local_path):
                status = TestStatus.FAIL
                break
        return status

    def _can_be_loaded(self, path: Path) -> bool:
        rdflib = self.import_module("rdflib")
        graph = rdflib.Graph()

        success = True
        with path.open("r") as infile:
            try:
                graph.parse(infile, format="json-ld")
            except Exception:
                success = False
        return success


class LibTiffInfoTest(ExternalTestMixin, TestABC):
    tier = 2

    def generate_process(self) -> Process:
        file = self.get_file()
        command_args = ["tiffinfo", file.local_path.as_posix()]
        process = Process(
            container="quay.io/sagebionetworks/libtiff:2.0",
            command_args=command_args,
        )
        return process


class BioFormatsInfoTest(ExternalTestMixin, TestABC):
    tier = 2

    def generate_process(self) -> Process:
        file = self.get_file()
        command_args = [
            "/opt/bftools/showinf",
            "-nopix",
            "-novalid",
            "-nocore",
            file.local_path.as_posix(),
        ]
        process = Process(
            container="quay.io/sagebionetworks/bftools:latest",
            command_args=command_args,
        )
        return process


class OmeXmlSchemaTest(ExternalTestMixin, TestABC):
    tier = 2

    def generate_process(self) -> Process:
        file = self.get_file()
        command_args = [
            "/opt/bftools/xmlvalid",
            file.local_path.as_posix(),
        ]
        process = Process(
            container="quay.io/sagebionetworks/bftools:latest",
            command_args=command_args,
        )
        return process



from dcqc.file import FileType
from dcqc.suites.suite_abc import SuiteABC
from dcqc.tests import tests


# TODO: Consider moving the filetype-test association logic
#       to the file types
class FileSuite(SuiteABC):
    file_type = FileType.get_file_type("*")
    add_tests = (tests.FileExtensionTest, tests.Md5ChecksumTest)


class JsonSuite(FileSuite):
    file_type = FileType.get_file_type("JSON")
    add_tests = (tests.JsonLoadTest,)


class JsonLdSuite(JsonSuite):
    file_type = FileType.get_file_type("JSON-LD")
    add_tests = (tests.JsonLdLoadTest,)


class TiffSuite(FileSuite):
    file_type = FileType.get_file_type("TIFF")
    add_tests = (tests.LibTiffInfoTest,)


class OmeTiffSuite(TiffSuite):
    file_type = FileType.get_file_type("OME-TIFF")
    add_tests = (tests.OmeXmlSchemaTest, tests.BioFormatsInfoTest)


class TSVSuite(FileSuite):
    file_type = FileType.get_file_type("TSV")
